#include <avr/io.h>
#include <util/twi.h>

#include "i2c.h"

// TWI mode
typedef enum _tagTWIMode{
	TWI_MODE_SEND,		// send
	TWI_MODE_RECEIVE,	// recv
}TWIMode;

// TWI ack
typedef enum _tagTWIACKMode{
	TWI_ACKMODE_ACK,	// ack
	TWI_ACKMODE_NACK,	// nack
}TWIACKMode;

// SCL clock = CPU_CLOCK / (16+2*(TWBR)*prescaler)
// TWBR = (CPU_CLOCK/SCL clock-16)/prescaler/2
// ex.
// CPU clock = 16MHz
// SCL clock = 400kHz
// Prescaler = 1
// TWBR = 12
void init_i2c(void){
	TWSR = 0; // 1/1
	//TWSR = _BV(TWPS0); // 1/4
	//TWSR = _BV(TWPS1); // 1/16
	//TWSR = _BV(TWPS0) | _BV(TWPS1); // 1/64
	TWBR = 12;
	TWCR = _BV(TWEN);
}

int8_t i2c_start(void){
	TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);

	loop_until_bit_is_set(TWCR, TWINT);

	if((TWSR & 0xF8) == TW_START){
		return TW_START;
	}else if((TWSR & 0xF8) == TW_REP_START){
		return TW_REP_START;
	}

	return  TW_BUS_ERROR;
}

int8_t i2c_setaddr(uint8_t addr, TWIMode mode){
	if(mode == TWI_MODE_SEND){
		TWDR = addr + TW_WRITE;
		TWCR = _BV(TWINT) | _BV(TWEN);
		loop_until_bit_is_set(TWCR, TWINT);

		if((TWSR & 0xF8) == TW_MT_SLA_ACK ){
			return TW_MT_SLA_ACK ;
		}else if((TWSR & 0xF8) == TW_MT_SLA_NACK){
			return TW_MT_SLA_NACK;
		}
	}else if(mode == TWI_MODE_RECEIVE){
		TWDR = addr + TW_READ;
		TWCR = _BV(TWINT) | _BV(TWEN);
		loop_until_bit_is_set(TWCR, TWINT);

		if((TWSR & 0xF8) == TW_MR_SLA_ACK){
			return TW_MR_SLA_ACK;
		} else if((TWSR & 0xF8) == TW_MR_SLA_NACK){
			return TW_MR_SLA_NACK;
		}
	}else{
		return TW_BUS_ERROR;
	}

	return TW_BUS_ERROR;
}

int8_t i2c_snddata(uint8_t data){
	TWDR = data;
	TWCR = _BV(TWINT) | _BV(TWEN);

	loop_until_bit_is_set(TWCR, TWINT);

	if((TWSR & 0xF8) == TW_MT_DATA_ACK){
		return TW_MT_DATA_ACK;
	}else if((TWSR & 0xF8) == TW_MT_DATA_NACK){
		return TW_MT_DATA_NACK;
	}

	return TW_BUS_ERROR;
}

int8_t i2c_rcvdata(uint8_t *data, TWIACKMode mode){
	if(TWI_ACKMODE_ACK == mode){
		TWCR = _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
	}else{
		TWCR = _BV(TWINT) | _BV(TWEN);
	}

	loop_until_bit_is_set(TWCR, TWINT);

	if((TWSR & 0xF8) == TW_MR_DATA_ACK){
		*data = TWDR;
		return TW_MR_DATA_ACK;
	}else if((TWSR & 0xF8) == TW_MR_DATA_NACK){
		*data = TWDR;
		return TW_MR_DATA_NACK;
	}

	return TW_BUS_ERROR;
}

void i2c_stop(void){
	TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);
}

void i2c_write(uint8_t id, uint8_t addr, uint8_t data){
	while(1){
		if(i2c_start() == TW_BUS_ERROR){
			continue;
		}
		if(i2c_setaddr(id, TWI_MODE_SEND) == TW_BUS_ERROR){
			continue;
		}
		if(i2c_snddata(addr) == TW_BUS_ERROR){
			continue;
		}
		if(i2c_snddata(data) == TW_BUS_ERROR){
			continue;
		}
		i2c_stop();
		break;
	}
}

void i2c_write_multi(uint8_t id, uint8_t start, uint8_t *data, uint8_t num){
	uint8_t	i = 0;

	while(1){
		if(i2c_start() == TW_BUS_ERROR){
			continue;
		}
		if(i2c_setaddr(id, TWI_MODE_SEND) == TW_BUS_ERROR){
			continue;
		}
		if(i2c_snddata(start) == TW_BUS_ERROR){
			continue;
		}
		for(i=0; i<num; i++){
			if(i2c_snddata(data[i]) == TW_BUS_ERROR) break;
		}
		if(num != i){
			continue;
		}
		i2c_stop();
		break;
	}
}

void i2c_read(uint8_t id, uint8_t addr, uint8_t *data){
	while(1){
		if(i2c_start() == TW_BUS_ERROR){
			continue;
		}
		if(i2c_setaddr(id, TWI_MODE_SEND) == TW_BUS_ERROR){
			continue;
		}
		if(i2c_snddata(addr) == TW_BUS_ERROR){
			continue;
		}
		if(i2c_start() == TW_BUS_ERROR){
			continue;
		}
		if(i2c_setaddr(id, TWI_MODE_RECEIVE) == TW_BUS_ERROR){
			continue;
		}
		if(i2c_rcvdata(data, TWI_ACKMODE_NACK) == TW_BUS_ERROR){
			continue;
		}
		i2c_stop();
		break;
	}
}


void i2c_read_multi(uint8_t id, uint8_t start, uint8_t *data, uint8_t num){
	uint8_t	i = 0;

	while(1){
		if(i2c_start() == TW_BUS_ERROR){
			continue;
		}
		if(i2c_setaddr(id, TWI_MODE_SEND) == TW_BUS_ERROR){
			continue;
		}
		if(i2c_snddata(start) == TW_BUS_ERROR){
			continue;
		}
		if(i2c_start() == TW_BUS_ERROR){
			continue;
		}
		if(i2c_setaddr(id, TWI_MODE_RECEIVE) == TW_BUS_ERROR){
			continue;
		}
		for(i=0; i<num-1; i++){
			if(i2c_rcvdata(&data[i], TWI_ACKMODE_ACK) == TW_BUS_ERROR) break;
		}
		if(num - 1 != i){
			continue;
		}
		if(i2c_rcvdata(&data[i], TWI_ACKMODE_NACK) == TW_BUS_ERROR){
			continue;
		}
		i2c_stop();
		break;
	}
}

