#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdio.h>
#include <string.h>

#include "command.h"
#include "common.h"
#include "DS3234S.h"
#include "eeprom.h"
#include "extern.h"
#include "LPS331.h"
#include "mode.h"
#include "uart.h"
#include "xitoa.h"

// 1-7 to Day (same as DS3234S)
char* get_day(uint8_t num){
	switch(num){
	case 1:
		return "Sun";
		break;

	case 2:
		return "Mon";
		break;

	case 3:
		return "Tue";
		break;

	case 4:
		return "Wed";
		break;

	case 5:
		return "Thu";
		break;

	case 6:
		return "Fri";
		break;

	case 7:
		return "Sat";
		break;

	default:
		return "INV";
		break;
	}

	return "INV";
}

void command_error_msg(uint8_t command, const uint8_t *str, uint8_t ret, uint8_t num){
	xprintf(PSTR("%c:Invalid command or number str=%s ret=%d num=%d \n"), command, str, ret, num);
}

// exec command by a-z A-Z, see command.h
void exec_command(const uint8_t *str){
	uint8_t command;
	int num;
	//int num_time[3];
	uint8_t num_time[3] = {0};
	int ret;
	//uint8_t buf[3] = {0};
	//uint8_t day;
	//uint8_t i;
	uint8_t temp_preset[7] = {0};
	uint8_t temp_photo_num[8] = {0};
	uint8_t temp_photo_Ldot[8] = {0};
	uint8_t temp_photo_Rdot[8] = {0};
	
	command = 0;
	num = 0;
	ret = 0;
	ret = sscanf((char *)str, "%c%d", &command, &num);

	switch(command){
// a

	case SET_AMPM_FLAG: // A
		if(ret != 2 || (num != FULL && num != HALF)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			change_hour_type(num);
		}
		break;

	case SET_SHUFFLE_CON: // b
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			if(num == ENABLE){
				start_loop_shuffle();
				do_shuffle();
			}else{
				stop_loop_shuffle();
				stop_shuffle();
			}
		}
		break;

// B

	case SET_CENTURY: // c
		if(ret != 2 || num < 0 || num > 99){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_century(num);
		}
		break;

	case SHOW_OP_TIME: // C
		show_op_time();
		break;

	case SET_DATE: // d
		if(ret != 2 || num <= 0 || num > 31){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			rtc_write_int(RTC_DATE_ADDR, num);
		}
		break;

	case SET_DAY: // D
		if(ret != 2 || num <= 0 || num > 7){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			rtc_write_int(RTC_DAY_ADDR, num);
		}
		break;

	case SET_CFADE: // e
		if(ret != 2 || num < XFADE_DIV_MIN || num > XFADE_DIV_MAX){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_xfade_div(num);
		}
		break;

// E

	case SET_SHUFFLE_TIME1: // f
		ret = sscanf((char *)str, "%c,%hhu,%hhu,%hhu", &command, &num_time[0], &num_time[1], &num_time[2]);
		if(ret != 4 || num_time[0] > 23 || num_time[1] > 59 || num_time[2] > 59 ||
		   num_time[0] < 0 || num_time[1] < 0 || num_time[2] < 0){
#ifdef DEBUG
			command_error_msg(command, str, ret, 0);
#endif
		}else{
			eep_update_shuffle_time_1(num_time[0], num_time[1], num_time[2]);
		}
		break;

	case SET_SHUFFLE_TIME2: // F
		ret = sscanf((char *)str, "%c,%hhu,%hhu,%hhu", &command, &num_time[0], &num_time[1], &num_time[2]);
		if(ret != 4 || num_time[0] > 23 || num_time[1] > 59 || num_time[2] > 59 ||
		   num_time[0] < 0 || num_time[1] < 0 || num_time[2] < 0){
#ifdef DEBUG
			command_error_msg(command, str, ret, 0);
#endif
		}else{
			eep_update_shuffle_time_2(num_time[0], num_time[1], num_time[2]);
		}
		break;

	case SET_GPS_LOG: // g
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_gps_log(num);
		}
		break;

	case USE_GPS: // G
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_use_gps(num);
		}
		break;

	case SET_HOUR: // h
		if(ret != 2 || num < 0 || num >= 24){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			rtc_write_int(RTC_HOUR_ADDR, num);
		}
		break;

	case USE_LIGHT_SENSOR_HVPS: // H
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_light_sensor_hvps(num);
		}
		break;

	case SET_SHUFFLE_PRESET:// i
		ret = sscanf((char *)str, "%c,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu",
									  &command, &temp_preset[0], &temp_preset[1], &temp_preset[2], &temp_preset[3],
									  &temp_preset[4], &temp_preset[5], &temp_preset[6]);
		if(ret != 8){
#ifdef DEBUG
			command_error_msg(command, str, ret, 0);
#endif
		}else{
			eep_update_shuffle_digit(temp_preset, sizeof(temp_preset));
		}
		break;

	case DO_SHUFFLE_PRESET: // I
		do_shuffle_preset();
		break;

	case RESET_GPS: // j
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			if(num == 1){reset_gps();}
		}
		break;

/*
	case GPS_PASSTHURU: // J
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
//			command_error_msg(command, str, ret, num);
		}else{
			if(num == TRUE){
				gps_passthuru = ENABLE;
			}else{
				gps_passthuru = DISABLE;
			}
		}
		break;
*/

	case SET_GPS_TIMEZONE: // k
		if(ret != 2 || num < GPS_TIMEZONE_MIN || num > GPS_TIMEZONE_MAX){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_gps_tz(num);
		}
		break;

// K

	case USE_LIGHT_SENSOR: // l
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_use_light_sensor(num);
		}
		break;

	case SET_LIGHT_SENSOR_TH: // L
		if(ret != 2 || num < LIGHT_SENSOR_TH_MIN || num > LIGHT_SENSOR_TH_MAX){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_light_sensor_th(num);
		}
		break;

	case SET_MIN: // m
		if(ret != 2 || num < 0 || num > 59){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			rtc_write_int(RTC_MIN_ADDR, num);
		}
		break;

	case SET_MONTH: // M
		if(ret != 2 || num <= 0 || num > 12){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			rtc_write_int(RTC_MONTH_ADDR, num);
		}
		break;

	case SET_NIXIE_PULSE: // n
		if(ret != 2 || num < DUTY_MIN || num > DUTY_MAX){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_duty(num);
		}
		break;

// N

	case SET_MODE: // o
		if(ret != 2 || num < MODE_TIME || num > MODE_TEST){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			if(num == CURRENT_MODE){
				if(num == MODE_PRESSURE){toggle_pres_type();}
				if(num == MODE_TEMPERATURE){toggle_temp_type();}
				if(num == MODE_TEST){change_test_pattern();}
			}
			stop_loop_shuffle();
			stop_shuffle();
			global_config_data.mode = num;
			update_mode(num);
		}
		break;

// O

	case SET_DOT_PATTERN: // p
		if(ret != 2 || num < 0 || num > DOT_OFF){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_dot_mode(num);
		}
		break;

	case SET_AMPM: // P
		if(ret != 2 || num < 0 || num > 1){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			change_hour_type(num);
		}
		break;

	case SET_PHOTO_NUM:// q
		if(strlen((char *)str) >= 49){
			ret = sscanf((char *)str, "%c,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu,%hhu",
										  &command, &temp_photo_num[0], &temp_photo_Ldot[0], &temp_photo_Rdot[0],
										  &temp_photo_num[1], &temp_photo_Ldot[1], &temp_photo_Rdot[1],
										  &temp_photo_num[2], &temp_photo_Ldot[2], &temp_photo_Rdot[2],
										  &temp_photo_num[3], &temp_photo_Ldot[3], &temp_photo_Rdot[3],
										  &temp_photo_num[4], &temp_photo_Ldot[4], &temp_photo_Rdot[4],
										  &temp_photo_num[5], &temp_photo_Ldot[5], &temp_photo_Rdot[5],
										  &temp_photo_num[6], &temp_photo_Ldot[6], &temp_photo_Rdot[6],
										  &temp_photo_num[7], &temp_photo_Ldot[7], &temp_photo_Rdot[7]);
			if(ret != 25){
#ifdef DEBUG
				command_error_msg(command, str, ret, 0);
#endif
			}else{
				eep_update_photo_digit(temp_photo_num, sizeof(temp_photo_num), temp_photo_Ldot, sizeof(temp_photo_Ldot), temp_photo_Rdot, sizeof(temp_photo_Rdot));
			}
			global_photo_mode_pos = 8;
			update_mode(MODE_PHOTO);
		}
		break;

// Q

	case SET_TURN_OFF_START: // r
		ret = sscanf((char *)str, "%c,%hhu,%hhu,%hhu", &command, &num_time[0], &num_time[1], &num_time[2]);
		if(ret != 4 || num_time[0] > 23 || num_time[1] > 59 || num_time[2] > 59 ||
		   num_time[0] < 0 || num_time[1] < 0 || num_time[2] < 0){
#ifdef DEBUG
			command_error_msg(command, str, ret, 0);
#endif
		}else{
			eep_update_turnoff_start(num_time[0], num_time[1], num_time[2]);
		}
		break;

	case SET_TURN_OFF_END: // R
		ret = sscanf((char *)str, "%c,%hhu,%hhu,%hhu", &command, &num_time[0], &num_time[1], &num_time[2]);
		if(ret != 4 || num_time[0] > 23 || num_time[1] > 59 || num_time[2] > 59 ||
		   num_time[0] < 0 || num_time[1] < 0 || num_time[2] < 0){
#ifdef DEBUG
			command_error_msg(command, str, ret, 0);
#endif
		}else{
			eep_update_turnoff_end(num_time[0], num_time[1], num_time[2]);
		}
		break;

	case SET_SEC: // s
		if(ret != 2 || num < 0 || num > 59){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			rtc_write_int(RTC_SEC_ADDR, num);
		}
		break;

	case USE_SHUFFLE: // S
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_use_shuffle(num);
		}
		break;

	case SET_TEMP_TYPE: // t
		if(ret != 2 || (num != CEL && num != FAH && num != KEL)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_temp_type(num);
		}
		break;

	case USE_TURNOFF: // T
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_use_turnoff(num);
		}
		break;

	case SET_PRES_TYPE: // u
		if(ret != 2 || (num != PA && num != MMHG)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_pres_type(num);
		}
		break;

// U

	case SET_HVPS: // v
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			if(num == 1){
				global_hvps_status |= HVPS_MAIN;
			}else if(num == 0){
				global_hvps_status &= (uint8_t)~HVPS_MAIN;
			}
		}
		break;

// V

	case READ_SERIAL_NO: // w
		xprintf(PSTR("SERIAL NO=%d \n"), global_config_data.serial);
		break;

	case WRITE_SERIAL_NO: // W
		if(ret != 2 || num < 0 || num > 0xFFFFFFFF){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_serial(num);
		}
		break;

	case USE_CFADE: // x
		if(ret != 2 || (num != ENABLE && num != DISABLE)){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			eep_update_use_xfade(num);
		}
		break;

	case INITIALIZE: // X
		reset_all_setting();
		break;

	case SET_YEAR: // y
		if(ret != 2 || num < 0 || num > 99){
#ifdef DEBUG
			command_error_msg(command, str, ret, num);
#endif
		}else{
			rtc_write_int(RTC_YEAR_ADDR, num);
		}
		break;

// Y

	case DO_SHUFFLE: // z
		do_shuffle();
		break;

	case READ_SETTING: // Z
		output_setting();
		break;

	default:
		// do nothing
		break;
	}
}

uint8_t gen_24hr(void){
	uint8_t temp_hour;

	// 12hr @ AM 12-01-02-03-04-05-06-07-08-09-10-11
	if(global_config_data.ampm == HALF && !(global_rtc_hour & RTC_PM_FLAG)){
		temp_hour = ((global_rtc_hour & 0b00010000) >> 4) * 10 + (global_rtc_hour & 0b00001111);
		if(temp_hour == 12){
			return 0;
		}else{
			return ((global_rtc_hour & 0b00010000) >> 4) * 10 + (global_rtc_hour & 0b00001111);
		}
	// 12hr @ PM 12-01-02-03-04-05-06-07-08-09-10-11
	}else if(global_config_data.ampm == HALF && (global_rtc_hour & RTC_PM_FLAG)){
		temp_hour = ((global_rtc_hour & 0b00010000) >> 4) * 10 + (global_rtc_hour & 0b00001111);
		if(temp_hour == 12){
			return ((global_rtc_hour & 0b00010000) >> 4) * 10 + (global_rtc_hour & 0b00001111);
		}else{
			return ((global_rtc_hour & 0b00010000) >> 4) * 10 + (global_rtc_hour & 0b00001111) + 12;
		}
	// 24hr
	}else{
		return ((global_rtc_hour & 0b00110000) >> 4) * 10 + (global_rtc_hour & 0b00001111);;
	}

	return 0;
}

void output_setting(void){
	uint8_t hour;
	uint8_t hour_str[3];

	// get hour
	hour = gen_24hr();
	sprintf((char *)hour_str, "%02d", hour);
	hour_str[0] -= 48;
	hour_str[1] -= 48;

	// start
	uart1_putn('*');
	uart1_put('\n');
	uart1_put('\n');
	uart1_put('\n');

	uart1_putn('A'); // AMPM
	if(global_config_data.ampm == HALF){
		uart1_putn('1');
	}else{
		uart1_putn('0');
	}
	uart1_putn('B'); // Century
	xprintf(PSTR("%d%d\n"), global_config_data.century >> 4, global_config_data.century & 0b00001111);
	uart1_putn('C'); // year
	xprintf(PSTR("%d%d\n"), global_rtc_year >> 4, global_rtc_year & 0b00001111);
	uart1_putn('D'); // month
	xprintf(PSTR("%d%d\n"), global_rtc_month >> 4, global_rtc_month & 0b00001111);
	uart1_putn('E'); // date
	xprintf(PSTR("%d%d\n"), global_rtc_date >> 4, global_rtc_date & 0b00001111);
	uart1_putn('F'); // day
	xprintf(PSTR("%d\n"), global_rtc_day);
	uart1_putn('G'); // hour
	xprintf(PSTR("%d%d\n"), hour_str[0], hour_str[1]);
	uart1_putn('H'); // min
	xprintf(PSTR("%d%d\n"), global_rtc_min >> 4, global_rtc_min & 0b00001111);
	uart1_putn('I'); // sec
	xprintf(PSTR("%d%d\n"), global_rtc_sec >> 4, global_rtc_sec & 0b00001111);
	uart1_putn('J'); // dot pattern
	xprintf(PSTR("%d\n"), global_config_data.dot_mode);
	uart1_putn('K'); // use cfade
	xprintf(PSTR("%d\n"), global_config_data.use_xfade);
	uart1_putn('L'); // nixie pulse
	xprintf(PSTR("%d\n"), global_config_data.duty);
	uart1_putn('M'); // cfade length
	xprintf(PSTR("%d\n"), global_config_data.xfade_div);
	uart1_putn('N'); // use light_sensor
	xprintf(PSTR("%d\n"), global_config_data.use_light_sensor);
	uart1_putn('O'); // use hvps off
	xprintf(PSTR("%d\n"), global_config_data.light_sensor_hvps);
	uart1_putn('P'); // use threshold
	xprintf(PSTR("%d\n"), global_config_data.light_sensor_th);
	uart1_putn('Q'); // shuffle1 hour
	xprintf(PSTR("%d\n"), global_config_data.shuffle_hour[0]);
	uart1_putn('R'); // shuffle1 min
	xprintf(PSTR("%d\n"), global_config_data.shuffle_min[0]);
	uart1_putn('S'); // shuffle1 sec
	xprintf(PSTR("%d\n"), global_config_data.shuffle_sec[0]);
	uart1_putn('T'); // shuffle2 hour
	xprintf(PSTR("%d\n"), global_config_data.shuffle_hour[1]);
	uart1_putn('U'); // shuffle2 min
	xprintf(PSTR("%d\n"), global_config_data.shuffle_min[1]);
	uart1_putn('V'); // shuffle2 sec
	xprintf(PSTR("%d\n"), global_config_data.shuffle_sec[1]);
	uart1_putn('W'); // use shuffle
	xprintf(PSTR("%d\n"), global_config_data.use_shuffle);
	uart1_putn('X'); // operation time
	xprintf(PSTR("%lu\n"), global_config_data.sec_count);
	uart1_putn('Y'); // use_gps
	xprintf(PSTR("%d\n"), global_config_data.use_gps);
	uart1_putn('Z'); // use gps log
	xprintf(PSTR("%d\n"), global_config_data.gps_log);

	uart1_putn('a'); // turnoff start hour
	xprintf(PSTR("%d\n"), global_config_data.turnoff_hour[0]);
	uart1_putn('b'); // turnoff start min
	xprintf(PSTR("%d\n"), global_config_data.turnoff_min[0]);
	uart1_putn('c'); // turnoff start sec
	xprintf(PSTR("%d\n"), global_config_data.turnoff_sec[0]);
	uart1_putn('d'); // turnoff end hour
	xprintf(PSTR("%d\n"), global_config_data.turnoff_hour[1]);
	uart1_putn('e'); // turnoff end min
	xprintf(PSTR("%d\n"), global_config_data.turnoff_min[1]);
	uart1_putn('f'); // turn off end sec
	xprintf(PSTR("%d\n"), global_config_data.turnoff_sec[1]);
	uart1_putn('g'); // use turnoff
	xprintf(PSTR("%d\n"), global_config_data.use_turnoff);
	uart1_putn('h'); // meter preset
	xprintf(PSTR("%d,%d,%d,%d,%d,%d,%d,%d\n"), 
		global_config_data.shuffle_digit[0], global_config_data.shuffle_digit[1], global_config_data.shuffle_digit[2], global_config_data.shuffle_digit[3],
		global_config_data.shuffle_digit[4], global_config_data.shuffle_digit[5], global_config_data.shuffle_digit[6], global_config_data.shuffle_digit[7]);
	uart1_putn('i'); // gps timezone
	xprintf(PSTR("%d\n"), global_config_data.gps_tz);

	uart1_put('\n');
	uart1_put('\n');
	uart1_put('\n');
	uart1_putn('&');
	// end
}

